-- == SRB2Randomizer: Base file ==
-- Basic initialization of the main variables.

rawset(_G, "randomizer", {})

randomizer.seed = nil
randomizer.rv = nil

addHook("NetVars", function(net)
	randomizer.seed = net(randomizer.seed)
	randomizer.rv = net(randomizer.rv)
end)

randomizer.cvars = {}

local function initSeed(input)
	local num = tonumber(input)
	local displaystring = false

	if (num != nil)
		randomizer.seed = abs(num)
	elseif (type(input) == "string")
		local newseed = 0

		for i = 1,#input
			local b = input:byte(i,i)
			newseed = $1 + (b << i)
		end

		newseed = abs($1)
		randomizer.seed = newseed
		displaystring = true
	else
		randomizer.seed = abs(P_RandomFixed() + P_RandomFixed() + P_RandomFixed() + P_RandomFixed() + P_RandomFixed() + P_RandomFixed() + P_RandomFixed() + P_RandomFixed())
	end

	local p = "SRB2Randomizer seed has been set to \130"..randomizer.seed.."\128."

	if (displaystring)
		p = $1.." (\"\130"..input.."\128\")"
	end

	print(p)
end

COM_AddCommand("randomizer_setseed", function(p, arg)
	initSeed(arg)

	if (server != nil)
		print("\3\131NOTICE:\128 Seed was changed mid-game! Restarting map...")
		G_ExitLevel(gamemap, true)
	end
end, 1)

addHook("MapChange", function()
	-- This is the "random value" used for all randomizations.
	-- This is changed from the default seed in any MapLoad effects,
	-- so that changing settings gives you different randomization,
	-- and generally make it feel a little more random without impacting seed consistency.
	-- Effects applied mid-game, such as characters, can use this value, but not modify it.
	-- Effects that are supposed to be consistent between different maps should just use the seed directly.
	randomizer.rv = randomizer.seed
end)

addHook("ThinkFrame", function()
	if (randomizer.seed == nil)
		if (leveltime == 2*TICRATE)
			-- Because of the way SRB2's RNG is initialized,
			-- providing a random seed when the script loads on the title screen would give you the same seed every time.
			-- That kinda defeats the point of a randomizer, so I put this message here.
			print("\3\131NOTICE:\128 A seed for SRB2Randomizer has not been set. Use the command 'randomizer_setseed' to set one.")
		end

		return
	end

	for p in players.iterate
		if (p.randomizer == nil)
			p.randomizer = {}
		end
	end
end)
