-- == SRB2Randomizer: Character randomization ==
-- Based off of toaster's character diceroll

randomizer.cvars.characters = CV_RegisterVar({
	name = "randomizer_characters",
	flags = CV_NETVAR|CV_SHOWMODIF,
	PossibleValue = {
		Off = 0,
		-- No character randomization.
		-- Use the character you pick at the start of the game.
		Replace = 1,
		-- Character randomization is based off of the global random seed.
		-- Each level gets it's own "forced skin" value, essentially.
		-- This is more fair for races.
		Random = 2,
		-- Character randomization is done using RNG per player spawn.
		-- You can get rid of a bad roll by retrying, or waste a good roll by dying.
	},
	defaultvalue = "Replace",
})

randomizer.cvars.noforcechar = CV_RegisterVar({
	name = "randomizer_noforcechar",
	flags = CV_NETVAR|CV_SHOWMODIF,
	PossibleValue = CV_OnOff,
	defaultvalue = "On",
	-- If characters are randomized, this option will ignore maps that force you to a character.
	-- This is "On" by default because the main example of its usage is SRB1 Remake.
	-- It can be disabled when playing a map pack that expects specific characters for stages. 
})

randomizer.cvars.prefcolors = CV_RegisterVar({
	name = "randomizer_prefcolors",
	flags = CV_SHOWMODIF,
	PossibleValue = CV_OnOff,
	defaultvalue = "Off",
	-- Choose whenever or not to randomize each character's prefcolor in SP/splitscreen mode.
})

local function betterSkinChange(player, num)
	R_SetPlayerSkin(player, num)
	COM_BufInsertText(player, "skin " + skins[num].name) -- This needs to be done to fix the MP character select menu.

	if not (netgame) -- Presumably, you'd want the skincolour to stay constant - since it can help identify a player, and would be a pain to set each map.
		local prefcolor = skins[num].prefcolor

		if (randomizer.cvars.prefcolors.value)
			prefcolor = ((skins[num].prefcolor + (num * 3) + randomizer.rv) % (MAXSKINCOLORS-1)) + 1
		end

		player.skincolor = prefcolor
		player.mo.color = prefcolor
		COM_BufInsertText(player, "color "..prefcolor) -- Ditto.
	end
end

addHook("ThinkFrame", function(mo)
	if (randomizer.seed == nil or randomizer.rv == nil)
		return
	end

	if (randomizer.cvars.characters.value != 1)
		return
	end

	for p in players.iterate
		if (p.mo and p.mo.valid)
			if (randomizer.cvars.noforcechar.value == 0)
			and (mapheaderinfo[gamemap].forcecharacter)
				p.randomizer.skin = #skins[mapheaderinfo[gamemap].forcecharacter]
			elseif (p.randomizer.skin == nil)
			or (p.randomizer.prevmap == nil)
			or (p.randomizer.prevmap != gamemap)
				local playernum = (#p+1)
				local randomvalue = abs(randomizer.rv * playernum + (gamemap * 3) - (gamemap>>1 / playernum<<1))
				p.randomizer.skin = (randomvalue % #skins)
			end

			p.randomizer.prevmap = gamemap

			if (#skins[p.mo.skin] != p.randomizer.skin)
				betterSkinChange(p, p.randomizer.skin)
			end
		end
	end
end)

addHook("PlayerSpawn", function(p)
	if (randomizer.seed == nil or randomizer.rv == nil)
		return
	end

	if (randomizer.cvars.characters.value != 2)
		return
	end

	if (randomizer.cvars.noforcechar.value == 0)
	and (mapheaderinfo[gamemap].forcecharacter)
		p.randomizer.skin = #skins[mapheaderinfo[gamemap].forcecharacter]
	else
		p.randomizer.skin = P_RandomKey(#skins)
	end

	betterSkinChange(p, p.randomizer.skin)
end)
