-- == SRB2Randomizer: Enemy randomization ==
-- Changes enemy types in the levels around

randomizer.cvars.enemies = CV_RegisterVar({
	name = "randomizer_enemies",
	flags = CV_NETVAR|CV_SHOWMODIF,
	PossibleValue = {
		Off = 0,
		-- No enemy randomization.
		-- Use the original level's enemy layout.
		Zone = 1,
		-- Enemies will be properly themed to the level you're in,
		-- by only using enemies used in the current level.
		-- This option has more cohesion.
		Random = 2,
		-- Enemies from anywhere in the game can appear.
		-- This option is a lot more chaotic, but allows for custom enemies to appear.
	},
	defaultvalue = "Zone",
})

local function replaceEnemies()
	local enemytypes = {}
	local enemyspots = {}

	if (randomizer.cvars.enemies.value == 1)
		for mo in thinkers.iterate("mobj")
			if not (mo and mo.valid)
				continue
			end

			if (mo.flags & MF_ENEMY)
				table.insert(enemyspots, mo)

				local nodupes = true
				if (#enemytypes > 0)
					for i = 1,#enemytypes
						if (enemytypes[i] == mo.type)
							nodupes = false
							break
						end
					end
				end

				if (nodupes)
					table.insert(enemytypes, mo.type)
				end
			end
		end
	elseif (randomizer.cvars.enemies.value == 2)
		for i = 0,#mobjinfo-1
			if (mobjinfo[i].flags & MF_ENEMY)
			and (mobjinfo[i].doomednum != -1) -- make sure it's intended to be placeable
				local addamount = 4

				-- Weighted amounts based on type
				if (i == MT_GOOMBA)
				or (i == MT_BLUEGOOMBA)
					-- Mario Mode enemies
					addamount = 2
				elseif (i == MT_SRB1_CRAWLA)
				or (i == MT_SRB1_BAT)
				or (i == MT_SRB1_ROBOFISH)
				or (i == MT_SRB1_VOLCANOGUY)
				or (i == MT_SRB1_HOPPY)
				or (i == MT_SRB1_HOPPYWATER)
				or (i == MT_SRB1_HOPPYSKYLAB)
				or (i == MT_SRB1_MMZFLYING)
				or (i == MT_SRB1_UFO)
				or (i == MT_SRB1_GRAYBOT)
				or (i == MT_SRB1_ROBOTOPOLIS)
				or (i == MT_SRB1_RBZBUZZ)
				or (i == MT_SRB1_ROBOTOPOLIS)
				or (i == MT_SRB1_RBZSPIKES)
				or (i == MT_SRB1_GOLDBOT)
				or (i == MT_SRB1_GENREX)
					-- SRB1 enemies
					addamount = 1
				end

				for j = 1,addamount
					table.insert(enemytypes, i)
				end
			end
		end

		for mo in thinkers.iterate("mobj")
			if not (mo and mo.valid)
				continue
			end

			if (mo.flags & MF_ENEMY)
				table.insert(enemyspots, mo)
			end
		end
	else
		assert(false, "Bad enemy randomizer option")
		return
	end

	if (#enemytypes <= 1)
		-- Just one enemy type
		return
	end

	for i = 1,#enemyspots
		if (enemyspots[i] and enemyspots[i].valid)
			local mo = enemyspots[i]

			-- Because of how crazy some enemy types can be, especially Lua ones,
			-- we're spawning new enemies then removing the old ones later

			local newtype = enemytypes[(randomizer.rv % #enemytypes) + 1]

			randomizer.rv = abs($1 + (mo.x >> FRACBITS) + (mo.y >> FRACBITS)
				+ (mo.z >> FRACBITS) - mo.sprite - (mo.ceilingz >> FRACBITS)
				+ (mo.floorz >> FRACBITS)
			)

			if (mo.type != newtype)
				local newflip = (mo.flags2 & MF2_OBJECTFLIP)
				local newz = mo.z
				if (newflip)
					newz = $1 + mo.height - mobjinfo[newtype].height
				end

				mo.flags = $1|MF_NOCLIPTHING|MF_NOTHINK

				if (mo.type == MT_YELLOWSHELL)
					-- Yellow Spring Shells are considered important for progression and normally drop their spring.
					local springz = mo.z
					if (newflip)
						springz = $1 + mo.height - mobjinfo[MT_YELLOWSPRING].height
					end

					local spring = P_SpawnMobj(mo.x, mo.y, springz, MT_YELLOWSPRING)
					if (newflip)
						spring.eflags = $1|MFE_VERTICALFLIP
						spring.flags2 = $1|MF2_OBJECTFLIP
					end
				end

				local newmo = P_SpawnMobj(mo.x, mo.y, newz, newtype)
				newmo.angle = mo.angle
				if (newflip)
					newmo.eflags = $1|MFE_VERTICALFLIP
					newmo.flags2 = $1|MF2_OBJECTFLIP
				end

				--[[
				if (mo.spawnpoint and mo.spawnpoint.valid)
					newmo.spawnpoint = mo.spawnpoint
					newmo.spawnpoint.type = mobjinfo[newmo.type].doomednum
					newmo.spawnpoint.mobj = newmo
				end
				--]]

				P_RemoveMobj(enemyspots[i])
			end
		end
	end
end

addHook("MapLoad", function()
	if (randomizer.seed == nil or randomizer.rv == nil)
		return
	end

	if (randomizer.cvars.enemies.value != 0)
		replaceEnemies()
	end
end)
