-- == SRB2Randomizer: Level order randomization ==
-- Changes map order of the whole campagin

local lastseed = nil

local function forceOrderUpdate()
	lastseed = nil
end

randomizer.cvars.levelorder = CV_RegisterVar({
	name = "randomizer_levelorder",
	flags = CV_NETVAR|CV_CALL|CV_SHOWMODIF,
	PossibleValue = {
		Off = 0,
		-- No level randomization.
		-- Use the standard level order.
		All = 1,
		-- The randomized level order will contain all compatible levels.
		NoSecrets = 2,
		-- The randomized level order will only use levels from the standard campaign.
		-- Removes Aerial Garden, Azure Temple, Pipe Towers, SRB1 Remake, Spring Hill, etc.
		-- Packs that use unlockables for progression (like SUGOI) will give you a tiny map pool with this option set.
		NoAlternative = 3,
		-- The randomized level order will use everything except "alternative" gametypes. (2D Mode and NiGHTS)
		-- Removes SRB1 Remake and Spring Hill.
	},
	defaultvalue = "NoAlternative",
	func = forceOrderUpdate,
})

--[[
randomizer.cvars.ssorder = CV_RegisterVar({
	name = "randomizer_ssorder",
	flags = CV_NETVAR|CV_SHOWMODIF,
	PossibleValue = CV_OnOff,
	defaultvalue = "On",
	func = forceOrderUpdate,
	-- Decide whenever or not to randomize Special Stage order.
	-- (Disabled because of hardcoded Special Stage progression interfering.)
})
--]]

local levellist = {}

addHook("MapChange", function(map)
	if (randomizer.seed == nil or randomizer.rv == nil)
		return
	end

	if (randomizer.cvars.levelorder.value == 0)
	--and (randomizer.cvars.ssorder.value == 0)
		return
	end

	if (lastseed != randomizer.seed)
		local availablelevels = {}
	
		-- Initialize the new level order.
		levellist = {}

		-- Just sort of assume the map you're on is the "starting" level, since we have no access to SPStage_Start.
		table.insert(levellist, map)

		local tol = TOL_SP
		if (multiplayer or netgame)
			tol = TOL_COOP
		end

		for i = 1,1035
			if (mapheaderinfo[i] == nil)
				continue
			end

			if (i == map)
				continue
			end

			if (mapheaderinfo[i].typeoflevel & tol)
			and (randomizer.cvars.levelorder.value != 0)
				if (randomizer.cvars.levelorder.value == 2)
					-- No secret maps
					if (mapheaderinfo[i].unlockrequired != -1)
						continue
					end
				elseif (randomizer.cvars.levelorder.value == 3)
					-- No "alternative gamemode" maps
					if (mapheaderinfo[i].typeoflevel & TOL_2D)
					or (mapheaderinfo[i].typeoflevel & TOL_NIGHTS)
						continue
					end
				end

				table.insert(availablelevels, i)
			end
		end

		local randomvalue = randomizer.seed
		local header = mapheaderinfo[map]
		randomvalue = abs($1 + header.typeoflevel + (header.skynum * header.weather) - (header.actnum * header.nextlevel) - (i * 37))

		while (#availablelevels > 0)
			local i = (randomvalue % #availablelevels) + 1
			header = mapheaderinfo[availablelevels[i]]
			table.insert(levellist, availablelevels[i])
			randomvalue = abs($1 + header.typeoflevel + (header.skynum * header.weather) - (header.actnum * header.nextlevel) - (i * 37))
			table.remove(availablelevels, i)
		end

		lastseed = randomizer.seed
	end
end)

addHook("ThinkFrame", function()
	if (#levellist > 0)
		local newnextlevel = nil

		for i = 1,#levellist
			if (gamemap == levellist[i])
				if (levellist[i+1] == nil)
					newnextlevel = 1102 -- Credits
				else
					newnextlevel = levellist[i+1]
				end
				break
			end
		end

		if (newnextlevel != nil)
			G_SetCustomExitVars(newnextlevel)
		end
	end
end)
