-- == SRB2Randomizer: Music randomization ==
-- Changes all music tracks around

randomizer.cvars.music = CV_RegisterVar({
	name = "randomizer_music",
	flags = CV_SHOWMODIF,
	PossibleValue = {
		Off = 0,
		-- No music randomization.
		-- Use the original music for each level.
		Replace = 1,
		-- Music will get a consistent replacement per seed.
		-- Ex: the boss music will get replaced with the same song each time.
		Random = 2,
		-- Roll RNG for music choices on every level load.
		-- You will hear a different music track every time you reload a level.
	},
	defaultvalue = "Replace",
})

-- Pretty much just a copy of https://wiki.srb2.org/wiki/List_of_music
local musiclist = {
	"MAP01M",
	"MAP02M",
	"MAP03M",
	"MAP04M",
	"MAP05M",
	"MAP06M",
	"MAP07M",
	"MAP08M",
	"MAP09M",
	"MAP10M",
	"MAP11M",
	"MAP12M",
	"MAP13M",
	"MAP14M",
	"MAP15M",
	"MAP16M",
	"MAP17M",
	"MAP18M",
	"MAP19M",
	"MAP20M", -- More of a drowning-style jingle than music, but might be funny
	"MAP21M",
	"MAP22M",
	"MAP23M",
	"MAP24M",
	"MAP25M",
	"MAP26M",
	"MAP27M",
	"MAP28M",

	"MAP29M",
	"MAP30M",
	"MAP31M",
	"MAP32M",
	--"MAP40M", -- No OGG version, and is just an alternate version of MAPA9M
	"MAP45M",

	"MAP50M",
	"MAP51M",

	"MAP86M",
	"MAP87M",
	"MAP88M",
	"MAP89M",
	"MAP90M",
	"MAP91M",
	"MAP92M",
	"MAP93M",
	"MAP94M",
	"MAP95M",
	"MAP96M",
	"MAP97M",
	"MAP98M",
	"MAP99M",

	"MAPA9M",

	"MAPAZM",
	"MAPB0M",
	"MAPB1M",
	"MAPB2M",
	"MAPB3M",
	"MAPB4M",
	"MAPB5M",
	"MAPB6M",
	"MAPB7M",
	"MAPB8M",
	"MAPB9M",

	"MAPCSM",
	"MAPEFM",
	"MAPZZM",

	"READ_M",
	"CONTSC",
	"CHRSEL",
	"CREDIT",
	"RACENT",

	-- Added digital versions to all the SRB1 tracks below, for extra variety!
	"MAPAAM",
	"MAPABM",
	"MAPACM",
	"MAPADM",
	"MAPAEM",
	"MAPAFM",
	"MAPAGM",
	"MAPAHM",
	"MAPAIM",
	"MAPAJM",
	"MAPAKM",
	"MAPALM",
	"MAPAMM",
	"MAPANM",
	"MAPAOM",
	"MAPAPM",
	"MAPAQM",
	"MAPARM",
	"MAPASM",
	"MAPATM",
	"MAPAUM",
	"MAPAVM",
}

local musicindex = nil
local realmusname = nil
local savedmusname = nil
local lastmap = nil

local function setRandomizedMusic()
	local randomvalue = randomizer.seed
	local v = 0

	for i = 1,#mapmusname
		local b = mapmusname:byte(i,i)
		local mul = i

		if (i == #mapmusname)
			mul = 1
		end

		if (i & 1)
			v = $1 - (b * i)
		else
			v = $1 + (b * i)
		end
	end

	if (v == 0)
		v = (randomizer.seed >> 4) + 1
	end

	randomvalue = abs((($1 + v) << v) / v)

	musicindex = (randomvalue % #musiclist) + 1
	savedmusname = musiclist[musicindex]

	-- New behavior
	if (randomizer.cvars.music.value != 0)
		for p in players.iterate do
			if (p.bot)
				continue
			end

			COM_BufInsertText(p, "tunes " + musiclist[musicindex])
		end
	end
end

addHook("MapChange", function(map)
	if (server == nil)
		lastmap = nil
	end

	if (map != lastmap) or (randomizer.cvars.music.value == 2)
		musicindex = nil
		savedmusname = nil
	end

	lastmap = map
	--setRandomizedMusic()
end)

addHook("ThinkFrame", function()
	if (randomizer.seed == nil or randomizer.rv == nil)
		return
	end

	if (randomizer.cvars.music.value == 0)
		return
	end

	if (musicindex == nil or savedmusname == nil or mapmusname != savedmusname)
		setRandomizedMusic()
	end
end)
