-- == SRB2Randomizer: NiGHTS randomization ==
-- Changes item layouts of NiGHTS stages around

randomizer.cvars.nightsitems = CV_RegisterVar({
	name = "randomizer_nightsitems",
	flags = CV_NETVAR|CV_SHOWMODIF,
	PossibleValue = {
		Off = 0,
		-- No NiGHTS item randomization.
		-- Use the original stage's item layout.
		-- This is the safest bet when playing on multi-mare stages.
		Replace = 1,
		-- Completely randomize NiGHTS item types.
		-- This option creates a lot more variety.
		Shuffle = 2,
		-- Shuffle existing NiGHTS item positions around.
		-- This option is much more tame and keeps level balance in-tact.
	},
	defaultvalue = "Replace",
})

randomizer.cvars.nightspowerups = CV_RegisterVar({
	name = "randomizer_nightspowerups",
	flags = CV_NETVAR|CV_SHOWMODIF,
	PossibleValue = CV_OnOff,
	defaultvalue = "On",
	-- Choose whenever or not to randomize NiGHTS powerups.
})

-- Hoop spawning stuff copied from the source code.
-- Currently doesn't work because hnext/hprev aren't cleaned up the same way as target/tracer,
-- which means Lua can't really do anything with them.
-- Hopefully 2.2 will fix this and let me add hoops to the random mix.
local function vectorMatrixMultiply(v, m)
	local ret = {}

	ret[0] = FixedMul(v[0], m[0][0]) + FixedMul(v[1], m[1][0]) + FixedMul(v[2], m[2][0]) + FixedMul(v[3], m[3][0])
	ret[1] = FixedMul(v[0], m[0][1]) + FixedMul(v[1], m[1][1]) + FixedMul(v[2], m[2][1]) + FixedMul(v[3], m[3][1])
	ret[2] = FixedMul(v[0], m[0][2]) + FixedMul(v[1], m[1][2]) + FixedMul(v[2], m[2][2]) + FixedMul(v[3], m[3][2])
	ret[3] = FixedMul(v[0], m[0][3]) + FixedMul(v[1], m[1][3]) + FixedMul(v[2], m[2][3]) + FixedMul(v[3], m[3][3])

	return ret
end

local function rotateXMatrix(fa)
	local ret = {
		[0] = {},
		[1] = {},
		[2] = {},
		[3] = {}
	}
	local cosrad = cos(fa)
	local sinrad = sin(fa)

	ret[0][0] = FRACUNIT
	ret[0][1] = 0
	ret[0][2] = 0
	ret[0][3] = 0

	ret[1][0] = 0
	ret[1][1] = cosrad
	ret[1][2] = sinrad
	ret[1][3] = 0

	ret[2][0] = 0
	ret[2][1] = -sinrad
	ret[2][2] = cosrad
	ret[2][3] = 0

	ret[3][0] = 0
	ret[3][1] = 0
	ret[3][2] = 0
	ret[3][3] = FRACUNIT

	return ret
end

local function rotateZMatrix(fa)
	local ret = {
		[0] = {},
		[1] = {},
		[2] = {},
		[3] = {}
	}
	local cosrad = cos(fa)
	local sinrad = sin(fa)

	ret[0][0] = cosrad
	ret[0][1] = sinrad
	ret[0][2] = 0
	ret[0][3] = 0

	ret[1][0] = -sinrad
	ret[1][1] = cosrad
	ret[1][2] = 0
	ret[1][3] = 0

	ret[2][0] = 0
	ret[2][1] = 0
	ret[2][2] = FRACUNIT
	ret[2][3] = 0

	ret[3][0] = 0
	ret[3][1] = 0
	ret[3][2] = 0
	ret[3][3] = FRACUNIT

	return ret
end

local function spawnNightsHoop(x, y, z, pitch, yaw, size)
	local nextmobj = nil

	local hoopcenter = P_SpawnMobj(x, y, z, MT_HOOPCENTER)
	hoopcenter.movedir = pitch
	hoopcenter.movecount = yaw

	local hoopsize = size
	local hoopplacement = hoopsize * (4*FRACUNIT)

	hoopcenter.extravalue1 = hoopsize
	hoopcenter.extravalue2 = hoopplacement / 12

	for i = 0,hoopsize-1
		local fa = i * (ANGLE_MAX / hoopsize)
		local v = {
			[0] = FixedMul(cos(fa), hoopplacement),
			[1] = 0,
			[2] = FixedMul(sin(fa), hoopplacement),
			[3] = FRACUNIT
		}

		v = vectorMatrixMultiply(v, rotateXMatrix(pitch))
		v = vectorMatrixMultiply(v, rotateZMatrix(yaw))

		local finalx = x + v[0]
		local finaly = y + v[1]
		local finalz = z + v[2]

		local mobj = P_SpawnMobj(finalx, finaly, finalz, MT_HOOP)
		if (maptol & TOL_XMAS)
			mobj.state = mobj.info.seestate + (i & 1)
		end

		mobj.z = $1 - (mobj.height/2)
		mobj.target = hoopcenter
		mobj.fuse = 0

		if (nextmobj != nil)
			mobj.hprev = nextmobj
			mobj.hprev.hnext = mobj
		end

		nextmobj = mobj
	end

	repeat
		if (hoopsize >= 32)
			hoopsize = $1 - 16
		else
			hoopsize = $1 / 2
		end

		hoopplacement = hoopsize * (4*FRACUNIT)

		for i = 0,hoopsize-1
			local fa = i * (ANGLE_MAX / hoopsize)
			local v = {
				[0] = FixedMul(cos(fa), hoopplacement),
				[1] = 0,
				[2] = FixedMul(sin(fa), hoopplacement),
				[3] = FRACUNIT
			}

			v = vectorMatrixMultiply(v, rotateXMatrix(pitch))
			v = vectorMatrixMultiply(v, rotateZMatrix(yaw))

			local finalx = x + v[0]
			local finaly = y + v[1]
			local finalz = z + v[2]

			local mobj = P_SpawnMobj(finalx, finaly, finalz, MT_HOOPCOLLIDE)
			mobj.z = $1 - (mobj.height/2)

			if (nextmobj != nil)
				mobj.hprev = nextmobj
				mobj.hprev.hnext = mobj
			end

			nextmobj = mobj
		end
	until not (hoopsize >= 8)

	return hoopcenter
end

local function replaceNightsItem(inputmobj, newtype)
	if (inputmobj.type == newtype)
		return inputmobj
	end

	local mt = nil
	local mo = nil

	if (inputmobj.spawnpoint and inputmobj.spawnpoint.valid)
		mt = inputmobj.spawnpoint
	end

	if (inputmobj.type == MT_HOOPCENTER)
	or (newtype == MT_HOOPCENTER)
		-- Create a new mobj entirely since hoops are a mess
		if (newtype == MT_HOOPCENTER)
			mo = spawnNightsHoop(inputmobj.x, inputmobj.y, inputmobj.z, 0, 0, 32)
		else
			mo = P_SpawnMobj(inputmobj.x, inputmobj.y, inputmobj.z, newtype)
		end

		if (inputmobj.hnext and inputmobj.hnext.valid)
			local hn = inputmobj.hnext
			while (hn and hn.valid)
				local cur = hn
				hn = cur.hnext

				cur.hnext = nil
				cur.hprev = nil
				P_RemoveMobj(cur)
			end
		end

		P_RemoveMobj(inputmobj)
	else
		inputmobj.type = newtype
		mo = inputmobj
	end

	assert((mo and mo.valid), "Bad NiGHTS item replacement")

	-- Update spawnpoint since it's used for item respawn
	if (mt != nil)
		if (mo.type == MT_HOOPCENTER)
			mt.type = 1713
			mt.angle = (((mo.movedir * 255) / 359) & 255) + ((((mo.movecount * 255) / 359) & 255) << 8) 
		else
			if (mo.type == MT_BLUEBALL)
				mt.type = mobjinfo[MT_RING].doomednum
			else
				mt.type = mobjinfo[mo.type].doomednum
			end
		end

		mt.mobj = mo
	end

	if (G_IsSpecialStage() and mo.type == MT_NIGHTSWING)
		mo.state = S_RING
	elseif (maptol & TOL_XMAS) and not (mo.type == MT_RING)
		mo.state = mobjinfo[mo.type].seestate
	else
		mo.state = mobjinfo[mo.type].spawnstate
	end

	return mo
end

local function shuffleNightsItems()
	local items = {}
	local itemtypes = {}

	for mo in thinkers.iterate("mobj")
		if not (mo and mo.valid)
			continue
		end

		if (mo.type == MT_NIGHTSWING)
		or (mo.type == MT_RING)
		or (mo.type == MT_BLUEBALL)
		--or (mo.type == MT_HOOPCENTER)
			table.insert(items, mo)
			table.insert(itemtypes, mo.type)
		end
	end

	if (#items > 1)
		for i = 1,#items
			if (#itemtypes <= 0)
				break
			end

			local typeindex = (randomizer.rv % #itemtypes) + 1
			local mo = replaceNightsItem(items[i], itemtypes[typeindex])
			table.remove(itemtypes, typeindex)

			randomizer.rv = abs($1 + (mo.x >> FRACBITS) + (mo.y >> FRACBITS)
				+ (mo.z >> FRACBITS) - mo.sprite - (mo.ceilingz >> FRACBITS)
				+ (mo.floorz >> FRACBITS)
			)
		end
	end
end

local function randomNightsItems()
	local itemtypes = {
		[1] = MT_RING,
		[2] = MT_NIGHTSWING,
		--[3] = MT_HOOPCENTER,
	}

	if (G_IsSpecialStage())
		itemtypes[1] = MT_BLUEBALL
	end

	local count = 0
	local requiredcount = 0
	local nonrequireditems = {}
	local requireditemtype = itemtypes[1]

	for mo in thinkers.iterate("mobj")
		if not (mo and mo.valid)
			continue
		end

		if (mo.type == MT_EGGCAPSULE)
			-- Ensure there are enough items in the map to finish the map
			requiredcount = $1 + mo.health
			continue
		end

		for i = 1,#itemtypes
			if (mo.type == itemtypes[i])
				local newmo = replaceNightsItem(mo, itemtypes[(randomizer.rv % #itemtypes) + 1])

				if (newmo.type == requireditemtype)
					count = $1+1
				else
					table.insert(nonrequireditems, newmo)
				end

				randomizer.rv = abs($1 + (mo.x >> FRACBITS) + (mo.y >> FRACBITS)
					+ (mo.z >> FRACBITS) - mo.sprite - (mo.ceilingz >> FRACBITS)
					+ (mo.floorz >> FRACBITS)
				)
				break
			end
		end
	end

	while ((count < requiredcount) and (#nonrequireditems > 0))
		local i = (randomizer.rv % #nonrequireditems) + 1
		local mo = replaceNightsItem(nonrequireditems[i], requireditemtype)
		count = $1+1
		table.remove(nonrequireditems, i)

		randomizer.rv = abs($1 + (mo.x >> FRACBITS) + (mo.y >> FRACBITS)
			+ (mo.z >> FRACBITS) - mo.sprite - (mo.ceilingz >> FRACBITS)
			+ (mo.floorz >> FRACBITS)
		)
	end
end

local function replaceNightsPowerups()
	local poweruptypes = {
		MT_NIGHTSSUPERLOOP,
		MT_NIGHTSDRILLREFILL,
		MT_NIGHTSHELPER,
		MT_NIGHTSEXTRATIME,
		MT_NIGHTSLINKFREEZE,
	}

	for mo in thinkers.iterate("mobj")
		if not (mo and mo.valid)
			continue
		end

		local useseestate = (mo.state == mo.info.seestate)

		for i = 1,#poweruptypes
			if (mo.type == poweruptypes[i])
				mo.type = poweruptypes[(randomizer.rv % #poweruptypes) + 1]

				if (mo.spawnpoint and mo.spawnpoint.valid)
					mo.spawnpoint.type = mobjinfo[mo.type].doomednum
				end

				if (useseestate)
					mo.state = mobjinfo[mo.type].seestate
				else
					mo.state = mobjinfo[mo.type].spawnstate
				end

				randomizer.rv = abs($1 + (mo.x >> FRACBITS) + (mo.y >> FRACBITS)
					+ (mo.z >> FRACBITS) - mo.sprite - (mo.ceilingz >> FRACBITS)
					+ (mo.floorz >> FRACBITS)
				)
				break
			end
		end
	end
end

addHook("MapLoad", function(map)
	if not (mapheaderinfo[map].typeoflevel & TOL_NIGHTS)
		return
	end

	if (randomizer.seed == nil or randomizer.rv == nil)
		return
	end

	if (randomizer.cvars.nightsitems.value == 1)
		randomNightsItems()
	elseif (randomizer.cvars.nightsitems.value == 2)
		shuffleNightsItems()
	end

	if (randomizer.cvars.nightspowerups.value == 1)
		replaceNightsPowerups()
	end
end)
