-- == SRB2Randomizer: Ring randomization ==
-- Places new rings in pre-defined layouts throughout the map

randomizer.cvars.rings = CV_RegisterVar({
	name = "randomizer_rings",
	flags = CV_NETVAR|CV_SHOWMODIF,
	PossibleValue = {
		Off = 0,
		-- No ring randomization.
		-- Use the original level's ring layout.
		On = 1,
		-- Rings will be randomly placed across maps.
		-- Ceiling rings are only placed if there's a gravity flip sector.
		NeverCeilings = 2,
		-- Rings will be randomly placed across maps.
		-- Ceiling rings will never be placed.
		AlwaysCeilings = 3,
		-- Rings will be randomly placed across maps.
		-- Ceiling rings will always be placed.
	},
	defaultvalue = "On",
})

-- Whoooole lots of definin', over here.
local randomringpatterns = {}

-- 3x1 row
table.insert(randomringpatterns, {
	[1] = {x = 0, y = 0, z = 0},
	[2] = {x = 64, y = 0, z = 0},
	[3] = {x = -64, y = 0, z = 0},
})

-- 4x1 row
table.insert(randomringpatterns, {
	[1] = {x = 32, y = 0, z = 0},
	[2] = {x = -32, y = 0, z = 0},
	[3] = {x = 96, y = 0, z = 0},
	[4] = {x = -96, y = 0, z = 0},
})

-- 5x1 row
table.insert(randomringpatterns, {
	[1] = {x = 0, y = 0, z = 0},
	[2] = {x = 64, y = 0, z = 0},
	[3] = {x = -64, y = 0, z = 0},
	[4] = {x = 128, y = 0, z = 0},
	[5] = {x = -128, y = 0, z = 0},
})

-- 6x1 row
table.insert(randomringpatterns, {
	[1] = {x = 32, y = 0, z = 0},
	[2] = {x = -32, y = 0, z = 0},
	[3] = {x = 96, y = 0, z = 0},
	[4] = {x = -96, y = 0, z = 0},
	[5] = {x = 160, y = 0, z = 0},
	[6] = {x = -160, y = 0, z = 0},
})

-- 7x1 row
table.insert(randomringpatterns, {
	[1] = {x = 0, y = 0, z = 0},
	[2] = {x = 64, y = 0, z = 0},
	[3] = {x = -64, y = 0, z = 0},
	[4] = {x = 128, y = 0, z = 0},
	[5] = {x = -128, y = 0, z = 0},
	[6] = {x = 192, y = 0, z = 0},
	[7] = {x = -192, y = 0, z = 0},
})

-- 8x1 row
table.insert(randomringpatterns, {
	[1] = {x = 32, y = 0, z = 0},
	[2] = {x = -32, y = 0, z = 0},
	[3] = {x = 96, y = 0, z = 0},
	[4] = {x = -96, y = 0, z = 0},
	[5] = {x = 160, y = 0, z = 0},
	[6] = {x = -160, y = 0, z = 0},
	[7] = {x = 224, y = 0, z = 0},
	[8] = {x = -224, y = 0, z = 0},
})

-- 10x1 row
table.insert(randomringpatterns, {
	[1] = {x = 32, y = 0, z = 0},
	[2] = {x = -32, y = 0, z = 0},
	[3] = {x = 96, y = 0, z = 0},
	[4] = {x = -96, y = 0, z = 0},
	[5] = {x = 160, y = 0, z = 0},
	[6] = {x = -160, y = 0, z = 0},
	[7] = {x = 224, y = 0, z = 0},
	[8] = {x = -224, y = 0, z = 0},
	[9] = {x = 288, y = 0, z = 0},
	[10] = {x = -288, y = 0, z = 0},
})

-- 3x2 row
table.insert(randomringpatterns, {
	[1] = {x = 0, y = 32, z = 0},
	[2] = {x = 64, y = 32, z = 0},
	[3] = {x = -64, y = 32, z = 0},
	[4] = {x = 0, y = -32, z = 0},
	[5] = {x = 64, y = -32, z = 0},
	[6] = {x = -64, y = -32, z = 0},
})

-- 5x2 row
table.insert(randomringpatterns, {
	[1] = {x = 0, y = 32, z = 0},
	[2] = {x = 64, y = 32, z = 0},
	[3] = {x = -64, y = 32, z = 0},
	[4] = {x = 128, y = 32, z = 0},
	[5] = {x = -128, y = 32, z = 0},
	[6] = {x = 0, y = -32, z = 0},
	[7] = {x = 64, y = -32, z = 0},
	[8] = {x = -64, y = -32, z = 0},
	[9] = {x = 128, y = -32, z = 0},
	[10] = {x = -128, y = -32, z = 0},
})

-- 7x2 row
table.insert(randomringpatterns, {
	[1] = {x = 0, y = 32, z = 0},
	[2] = {x = 64, y = 32, z = 0},
	[3] = {x = -64, y = 32, z = 0},
	[4] = {x = 128, y = 32, z = 0},
	[5] = {x = -128, y = 32, z = 0},
	[6] = {x = 192, y = 32, z = 0},
	[7] = {x = -192, y = 32, z = 0},
	[8] = {x = 0, y = -32, z = 0},
	[9] = {x = 64, y = -32, z = 0},
	[10] = {x = -64, y = -32, z = 0},
	[11] = {x = 128, y = -32, z = 0},
	[12] = {x = -128, y = -32, z = 0},
	[13] = {x = 192, y = -32, z = 0},
	[14] = {x = -192, y = -32, z = 0},
})

-- 10x2 row
table.insert(randomringpatterns, {
	[1] = {x = 32, y = 32, z = 0},
	[2] = {x = -32, y = 32, z = 0},
	[3] = {x = 96, y = 32, z = 0},
	[4] = {x = -96, y = 32, z = 0},
	[5] = {x = 160, y = 32, z = 0},
	[6] = {x = -160, y = 32, z = 0},
	[7] = {x = 224, y = 32, z = 0},
	[8] = {x = -224, y = 32, z = 0},
	[9] = {x = 288, y = 32, z = 0},
	[10] = {x = 288, y = 32, z = 0},
	[11] = {x = 32, y = -32, z = 0},
	[12] = {x = -32, y = -32, z = 0},
	[13] = {x = 96, y = -32, z = 0},
	[14] = {x = -96, y = -32, z = 0},
	[15] = {x = 160, y = -32, z = 0},
	[16] = {x = -160, y = -32, z = 0},
	[17] = {x = 224, y = -32, z = 0},
	[18] = {x = -224, y = -32, z = 0},
	[19] = {x = 288, y = -32, z = 0},
	[20] = {x = -288, y = -32, z = 0},
})

-- 5x3 row
table.insert(randomringpatterns, {
	[1] = {x = 0, y = 0, z = 0},
	[2] = {x = 64, y = 0, z = 0},
	[3] = {x = -64, y = 0, z = 0},
	[4] = {x = 128, y = 0, z = 0},
	[5] = {x = -128, y = 0, z = 0},
	[6] = {x = 0, y = 64, z = 0},
	[7] = {x = 64, y = 64, z = 0},
	[8] = {x = -64, y = 64, z = 0},
	[9] = {x = 128, y = 64, z = 0},
	[10] = {x = -128, y = 64, z = 0},
	[11] = {x = 0, y = -64, z = 0},
	[12] = {x = 64, y = -64, z = 0},
	[13] = {x = -64, y = -64, z = 0},
	[14] = {x = 128, y = -64, z = 0},
	[15] = {x = -128, y = -64, z = 0},
})

-- 7x3 row
table.insert(randomringpatterns, {
	[1] = {x = 0, y = 0, z = 0},
	[2] = {x = 64, y = 0, z = 0},
	[3] = {x = -64, y = 0, z = 0},
	[4] = {x = 128, y = 0, z = 0},
	[5] = {x = -128, y = 0, z = 0},
	[6] = {x = 192, y = 0, z = 0},
	[7] = {x = -192, y = 0, z = 0},
	[8] = {x = 0, y = 64, z = 0},
	[9] = {x = 64, y = 64, z = 0},
	[10] = {x = -64, y = 64, z = 0},
	[11] = {x = 128, y = 64, z = 0},
	[12] = {x = -128, y = 64, z = 0},
	[13] = {x = 192, y = 64, z = 0},
	[14] = {x = -192, y = 64, z = 0},
	[15] = {x = 0, y = 64, z = 0},
	[16] = {x = 64, y = 64, z = 0},
	[17] = {x = -64, y = 64, z = 0},
	[18] = {x = 128, y = 64, z = 0},
	[19] = {x = -128, y = 64, z = 0},
	[20] = {x = 192, y = 64, z = 0},
	[21] = {x = -192, y = 64, z = 0},
})

-- 10x3 row
table.insert(randomringpatterns, {
	[1] = {x = 32, y = 0, z = 0},
	[2] = {x = -32, y = 0, z = 0},
	[3] = {x = 96, y = 0, z = 0},
	[4] = {x = -96, y = 0, z = 0},
	[5] = {x = 160, y = 0, z = 0},
	[6] = {x = -160, y = 0, z = 0},
	[7] = {x = 224, y = 0, z = 0},
	[8] = {x = -224, y = 0, z = 0},
	[9] = {x = 288, y = 0, z = 0},
	[10] = {x = -288, y = 0, z = 0},
	[11] = {x = 32, y = 64, z = 0},
	[12] = {x = -32, y = 64, z = 0},
	[13] = {x = 96, y = 64, z = 0},
	[14] = {x = -96, y = 64, z = 0},
	[15] = {x = 160, y = 64, z = 0},
	[16] = {x = -160, y = 64, z = 0},
	[17] = {x = 224, y = 64, z = 0},
	[18] = {x = -224, y = 64, z = 0},
	[19] = {x = 288, y = 64, z = 0},
	[20] = {x = -288, y = 64, z = 0},
	[21] = {x = 32, y = -64, z = 0},
	[22] = {x = -32, y = -64, z = 0},
	[23] = {x = 96, y = -64, z = 0},
	[24] = {x = -96, y = -64, z = 0},
	[25] = {x = 160, y = -64, z = 0},
	[26] = {x = -160, y = -64, z = 0},
	[27] = {x = 224, y = -64, z = 0},
	[28] = {x = -224, y = -64, z = 0},
	[29] = {x = 288, y = -64, z = 0},
	[30] = {x = -288, y = -64, z = 0},
})

-- 8-ring circle
table.insert(randomringpatterns, {
	[1] = {x = 64, y = 0, z = 0},
	[2] = {x = 48, y = 48, z = 0},
	[3] = {x = 0, y = 64, z = 0},
	[4] = {x = -48, y = 48, z = 0},
	[5] = {x = -64, y = 0, z = 0},
	[6] = {x = -48, y = -48, z = 0},
	[7] = {x = 0, y = -64, z = 0},
	[8] = {x = 48, y = -48, z = 0},
})

-- 16-ring circle
table.insert(randomringpatterns, {
	[1] = {x = 128, y = 0, z = 0},
	[2] = {x = 120, y = 24, z = 0},
	[3] = {x = 96, y = 96, z = 0},
	[4] = {x = 24, y = 120, z = 0},
	[5] = {x = 0, y = 128, z = 0},
	[6] = {x = -24, y = 120, z = 0},
	[7] = {x = -96, y = 96, z = 0},
	[8] = {x = -120, y = 8, z = 0},
	[1] = {x = -128, y = 0, z = 0},
	[2] = {x = -120, y = -24, z = 0},
	[3] = {x = -96, y = -96, z = 0},
	[4] = {x = -24, y = -120, z = 0},
	[5] = {x = 0, y = -128, z = 0},
	[6] = {x = 8, y = -120, z = 0},
	[7] = {x = 96, y = -96, z = 0},
	[8] = {x = 120, y = -24, z = 0},
})

-- 2x2 square
table.insert(randomringpatterns, {
	[1] = {x = 32, y = 32, z = 0},
	[2] = {x = 32, y = -32, z = 0},
	[3] = {x = -32, y = 32, z = 0},
	[4] = {x = -32, y = -32, z = 0},
})

-- 3x3 square
table.insert(randomringpatterns, {
	[1] = {x = 0, y = 0, z = 0},
	[2] = {x = 64, y = 0, z = 0},
	[3] = {x = -64, y = 0, z = 0},
	[4] = {x = 0, y = 64, z = 0},
	[5] = {x = 64, y = 64, z = 0},
	[6] = {x = -64, y = 64, z = 0},
	[7] = {x = 0, y = -64, z = 0},
	[8] = {x = 64, y = -64, z = 0},
	[9] = {x = -64, y = -4, z = 0},
})

-- 4x4 square
table.insert(randomringpatterns, {
	[1] = {x = 32, y = 32, z = 0},
	[2] = {x = 32, y = -32, z = 0},
	[3] = {x = 32, y = 96, z = 0},
	[4] = {x = 32, y = -96, z = 0},
	[5] = {x = -32, y = 32, z = 0},
	[6] = {x = -32, y = -32, z = 0},
	[7] = {x = -32, y = 96, z = 0},
	[8] = {x = -32, y = -96, z = 0},
	[9] = {x = 96, y = 32, z = 0},
	[10] = {x = 96, y = -32, z = 0},
	[11] = {x = 96, y = 96, z = 0},
	[12] = {x = 96, y = -96, z = 0},
	[13] = {x = -96, y = 32, z = 0},
	[14] = {x = -96, y = -32, z = 0},
	[15] = {x = -96, y = 96, z = 0},
	[16] = {x = -96, y = -96, z = 0},
})

-- 5x5 square
table.insert(randomringpatterns, {
	[1] = {x = 0, y = 0, z = 0},
	[2] = {x = 64, y = 0, z = 0},
	[3] = {x = -64, y = 0, z = 0},
	[4] = {x = 128, y = 0, z = 0},
	[5] = {x = -128, y = 0, z = 0},
	[6] = {x = 0, y = 64, z = 0},
	[7] = {x = 64, y = 64, z = 0},
	[8] = {x = -64, y = 64, z = 0},
	[9] = {x = 128, y = 64, z = 0},
	[10] = {x = -128, y = 64, z = 0},
	[11] = {x = 0, y = -64, z = 0},
	[12] = {x = 64, y = -64, z = 0},
	[13] = {x = -64, y = -64, z = 0},
	[14] = {x = 128, y = -64, z = 0},
	[15] = {x = -128, y = -64, z = 0},
	[16] = {x = 0, y = 128, z = 0},
	[17] = {x = 64, y = 128, z = 0},
	[18] = {x = -64, y = 128, z = 0},
	[19] = {x = 128, y = 128, z = 0},
	[20] = {x = -128, y = 128, z = 0},
	[21] = {x = 0, y = -128, z = 0},
	[22] = {x = 64, y = -128, z = 0},
	[23] = {x = -64, y = -128, z = 0},
	[24] = {x = 128, y = -128, z = 0},
	[25] = {x = -128, y = -128, z = 0},
})

-- 3x3 empty square
table.insert(randomringpatterns, {
	[1] = {x = 0, y = 64, z = 0},
	[2] = {x = 0, y = -64, z = 0},
	[3] = {x = 64, y = 0, z = 0},
	[4] = {x = 64, y = 64, z = 0},
	[5] = {x = 64, y = -64, z = 0},
	[6] = {x = -64, y = 0, z = 0},
	[7] = {x = -64, y = 64, z = 0},
	[8] = {x = -64, y = -64, z = 0},
})

-- 4x4 empty square
table.insert(randomringpatterns, {
	[1] = {x = 96, y = 96, z = 0},
	[2] = {x = 96, y = 32, z = 0},
	[3] = {x = 96, y = -32, z = 0},
	[4] = {x = 96, y = -96, z = 0},
	[5] = {x = -96, y = 96, z = 0},
	[6] = {x = -96, y = 32, z = 0},
	[7] = {x = -96, y = -32, z = 0},
	[8] = {x = -96, y = -96, z = 0},
	[9] = {x = 32, y = 96, z = 0},
	[10] = {x = 32, y = -96, z = 0},
	[11] = {x = -32, y = 96, z = 0},
	[12] = {x = -32, y = -96, z = 0},
})

-- 5x5 empty square
table.insert(randomringpatterns, {
	[1] = {x = 128, y = 0, z = 0},
	[2] = {x = 128, y = 64, z = 0},
	[3] = {x = 128, y = -64, z = 0},
	[4] = {x = 128, y = 128, z = 0},
	[5] = {x = 128, y = -128, z = 0},
	[6] = {x = -128, y = 0, z = 0},
	[7] = {x = -128, y = 64, z = 0},
	[8] = {x = -128, y = -64, z = 0},
	[9] = {x = -128, y = 128, z = 0},
	[10] = {x = -128, y = -128, z = 0},
	[11] = {x = 0, y = 128, z = 0},
	[12] = {x = 0, y = -128, z = 0},
	[13] = {x = 64, y = 128, z = 0},
	[14] = {x = 64, y = -128, z = 0},
	[15] = {x = -64, y = 128, z = 0},
	[16] = {x = -64, y = -128, z = 0},
})

-- 3x3 hexagon
table.insert(randomringpatterns, {
	[1] = {x = 0, y = 0, z = 0},
	[2] = {x = 64, y = 0, z = 0},
	[3] = {x = -64, y = 0, z = 0},
	[4] = {x = -32, y = 64, z = 0},
	[5] = {x = 32, y = 64, z = 0},
	[6] = {x = -32, y = -64, z = 0},
	[7] = {x = 32, y = -64, z = 0},
})

-- 5x3 hexagon
table.insert(randomringpatterns, {
	[1] = {x = 0, y = 0, z = 0},
	[2] = {x = 64, y = 0, z = 0},
	[3] = {x = -64, y = 0, z = 0},
	[4] = {x = 128, y = 0, z = 0},
	[5] = {x = -128, y = 0, z = 0},
	[6] = {x = -32, y = 64, z = 0},
	[7] = {x = 32, y = 64, z = 0},
	[8] = {x = -96, y = 64, z = 0},
	[9] = {x = 96, y = 64, z = 0},
	[10] = {x = -32, y = -64, z = 0},
	[11] = {x = 32, y = -64, z = 0},
	[12] = {x = -96, y = -64, z = 0},
	[13] = {x = 96, y = -64, z = 0},
})

-- 5x5 hexagon
table.insert(randomringpatterns, {
	[1] = {x = 0, y = 0, z = 0},
	[2] = {x = 64, y = 0, z = 0},
	[3] = {x = -64, y = 0, z = 0},
	[4] = {x = 128, y = 0, z = 0},
	[5] = {x = -128, y = 0, z = 0},
	[6] = {x = -32, y = 64, z = 0},
	[7] = {x = 32, y = 64, z = 0},
	[8] = {x = -96, y = 64, z = 0},
	[9] = {x = 96, y = 64, z = 0},
	[10] = {x = -32, y = -64, z = 0},
	[11] = {x = 32, y = -64, z = 0},
	[12] = {x = -96, y = -64, z = 0},
	[13] = {x = 96, y = -64, z = 0},
	[14] = {x = 0, y = 128, z = 0},
	[15] = {x = 64, y = 128, z = 0},
	[16] = {x = -64, y = 128, z = 0},
	[17] = {x = 0, y = -128, z = 0},
	[18] = {x = 64, y = -128, z = 0},
	[19] = {x = -64, y = -128, z = 0},
})

-- 2x2 X
table.insert(randomringpatterns, {
	[1] = {x = 0, y = 0, z = 0},
	[2] = {x = 64, y = 64, z = 0},
	[3] = {x = -64, y = 64, z = 0},
	[4] = {x = 64, y = -64, z = 0},
	[5] = {x = -64, y = -64, z = 0},
})

-- 3x3 X
table.insert(randomringpatterns, {
	[1] = {x = 0, y = 0, z = 0},
	[2] = {x = 64, y = 64, z = 0},
	[3] = {x = -64, y = 64, z = 0},
	[4] = {x = 64, y = -64, z = 0},
	[5] = {x = -64, y = -64, z = 0},
	[6] = {x = 128, y = 128, z = 0},
	[7] = {x = -128, y = 128, z = 0},
	[8] = {x = 128, y = -128, z = 0},
	[9] = {x = -128, y = -128, z = 0},
})

-- 4x4 X
table.insert(randomringpatterns, {
	[1] = {x = 0, y = 0, z = 0},
	[2] = {x = 64, y = 64, z = 0},
	[3] = {x = -64, y = 64, z = 0},
	[4] = {x = 64, y = -64, z = 0},
	[5] = {x = -64, y = -64, z = 0},
	[6] = {x = 128, y = 128, z = 0},
	[7] = {x = -128, y = 128, z = 0},
	[8] = {x = 128, y = -128, z = 0},
	[9] = {x = -128, y = -128, z = 0},
	[10] = {x = 192, y = 192, z = 0},
	[11] = {x = -192, y = 192, z = 0},
	[12] = {x = 192, y = -192, z = 0},
	[13] = {x = -192, y = -192, z = 0},
})

-- 3x3 triangle
table.insert(randomringpatterns, {
	[1] = {x = 0, y = 0, z = 0},
	[2] = {x = 64, y = 0, z = 0},
	[3] = {x = -64, y = 0, z = 0},
	[4] = {x = 0, y = 64, z = 0},
	[5] = {x = 0, y = -64, z = 0},
	[6] = {x = 64, y = -64, z = 0},
	[7] = {x = -64, y = -64, z = 0},
	[8] = {x = 128, y = -64, z = 0},
	[9] = {x = -128, y = -64, z = 0},
})

-- 5x5 triangle
table.insert(randomringpatterns, {
	[1] = {x = 0, y = 0, z = 0},
	[2] = {x = 64, y = 0, z = 0},
	[3] = {x = -64, y = 0, z = 0},
	[4] = {x = 128, y = 0, z = 0},
	[5] = {x = -128, y = 0, z = 0},
	[6] = {x = 0, y = 64, z = 0},
	[7] = {x = 64, y = 64, z = 0},
	[8] = {x = -64, y = 64, z = 0},
	[9] = {x = 0, y = 128, z = 0},
	[10] = {x = 0, y = -64, z = 0},
	[11] = {x = 64, y = -64, z = 0},
	[12] = {x = -64, y = -64, z = 0},
	[13] = {x = 128, y = -64, z = 0},
	[14] = {x = -128, y = -64, z = 0},
	[15] = {x = 192, y = -64, z = 0},
	[16] = {x = -192, y = -64, z = 0},
	[17] = {x = 0, y = -128, z = 0},
	[18] = {x = 64, y = -128, z = 0},
	[19] = {x = -64, y = -128, z = 0},
	[20] = {x = 128, y = -128, z = 0},
	[21] = {x = -128, y = -128, z = 0},
	[22] = {x = 192, y = -128, z = 0},
	[23] = {x = -192, y = -128, z = 0},
	[24] = {x = 256, y = -128, z = 0},
	[25] = {x = -256, y = -128, z = 0},
})

-- Small cluster
table.insert(randomringpatterns, {
	[1] = {x = 0, y = 0, z = 0},
	[2] = {x = 64, y = 0, z = 0},
	[3] = {x = -64, y = 0, z = 0},
	[4] = {x = 0, y = 64, z = 0},
	[5] = {x = 0, y = -64, z = 0},
	[6] = {x = 0, y = 0, z = 48},
})

-- Large cluster
table.insert(randomringpatterns, {
	[1] = {x = 0, y = 0, z = 0},
	[2] = {x = 64, y = 64, z = 0},
	[3] = {x = 64, y = -64, z = 0},
	[4] = {x = -64, y = 64, z = 0},
	[5] = {x = -64, y = -64, z = 0},
	[6] = {x = 0, y = 64, z = 48},
	[7] = {x = 0, y = -64, z = 48},
	[8] = {x = 64, y = 64, z = 48},
	[9] = {x = 64, y = -64, z = 48},
	[10] = {x = -64, y = 64, z = 48},
	[11] = {x = -64, y = -64, z = 48},
	[12] = {x = 0, y = 0, z = 96},
	[13] = {x = 0, y = 64, z = 96},
	[14] = {x = 0, y = -64, z = 96},
	[15] = {x = 64, y = 0, z = 96},
	[16] = {x = -64, y = 0, z = 96},
})

-- x1 snake A
table.insert(randomringpatterns, {
	[1] = {x = -64, y = 256, z = 0},
	[2] = {x = -48, y = 192, z = 0},
	[3] = {x = 0, y = 128, z = 0},
	[4] = {x = 48, y = 64, z = 0},
	[5] = {x = 64, y = 0, z = 0},
	[6] = {x = 48, y = -64, z = 0},
	[7] = {x = 0, y = -128, z = 0},
	[8] = {x = -48, y = -192, z = 0},
	[9] = {x = -64, y = -256, z = 0},
})

-- x1 snake B
table.insert(randomringpatterns, {
	[1] = {x = 64, y = 256, z = 0},
	[2] = {x = 48, y = 192, z = 0},
	[3] = {x = 0, y = 128, z = 0},
	[4] = {x = -48, y = 64, z = 0},
	[5] = {x = -64, y = 0, z = 0},
	[6] = {x = -48, y = -64, z = 0},
	[7] = {x = 0, y = -128, z = 0},
	[8] = {x = 48, y = -192, z = 0},
	[9] = {x = 64, y = -256, z = 0},
})

-- x2 snake A
table.insert(randomringpatterns, {
	[1] = {x = -96, y = 256, z = 0},
	[2] = {x = -80, y = 192, z = 0},
	[3] = {x = -32, y = 128, z = 0},
	[4] = {x = 16, y = 64, z = 0},
	[5] = {x = 32, y = 0, z = 0},
	[6] = {x = 16, y = -64, z = 0},
	[7] = {x = -32, y = -128, z = 0},
	[8] = {x = -80, y = -192, z = 0},
	[9] = {x = -96, y = -256, z = 0},
	[10] = {x = -32, y = 256, z = 0},
	[11] = {x = -16, y = 192, z = 0},
	[12] = {x = 32, y = 128, z = 0},
	[13] = {x = 80, y = 64, z = 0},
	[14] = {x = 96, y = 0, z = 0},
	[15] = {x = 80, y = -64, z = 0},
	[16] = {x = 32, y = -128, z = 0},
	[17] = {x = -16, y = -192, z = 0},
	[18] = {x = -32, y = -256, z = 0},
})

-- x2 snake B
table.insert(randomringpatterns, {
	[1] = {x = 32, y = 256, z = 0},
	[2] = {x = 16, y = 192, z = 0},
	[3] = {x = -32, y = 128, z = 0},
	[4] = {x = -80, y = 64, z = 0},
	[5] = {x = -96, y = 0, z = 0},
	[6] = {x = -80, y = -64, z = 0},
	[7] = {x = -32, y = -128, z = 0},
	[8] = {x = 16, y = -192, z = 0},
	[9] = {x = 32, y = -256, z = 0},
	[10] = {x = 96, y = 256, z = 0},
	[11] = {x = 80, y = 192, z = 0},
	[12] = {x = 32, y = 128, z = 0},
	[13] = {x = -16, y = 64, z = 0},
	[14] = {x = -32, y = 0, z = 0},
	[15] = {x = -16, y = -64, z = 0},
	[16] = {x = 32, y = -128, z = 0},
	[17] = {x = 80, y = -192, z = 0},
	[18] = {x = 96, y = -256, z = 0},
})

-- 2x2 elbow
table.insert(randomringpatterns, {
	[1] = {x = 32, y = 32, z = 0},
	[2] = {x = 32, y = -32, z = 0},
	[3] = {x = -32, y = 32, z = 0},
})

-- 3x3 elbow
table.insert(randomringpatterns, {
	[1] = {x = 64, y = 64, z = 0},
	[2] = {x = 64, y = 0, z = 0},
	[3] = {x = 64, y = -64, z = 0},
	[4] = {x = 0, y = -64, z = 0},
	[5] = {x = -64, y = -64, z = 0},
})

-- 4x4 elbow
table.insert(randomringpatterns, {
	[1] = {x = 96, y = 96, z = 0},
	[2] = {x = 96, y = 32, z = 0},
	[3] = {x = 96, y = -32, z = 0},
	[4] = {x = 96, y = -96, z = 0},
	[5] = {x = 32, y = -96, z = 0},
	[6] = {x = -32, y = -96, z = 0},
	[7] = {x = -96, y = -96, z = 0},
})

-- 5x5 elbow
table.insert(randomringpatterns, {
	[1] = {x = 128, y = 128, z = 0},
	[2] = {x = 128, y = 64, z = 0},
	[3] = {x = 128, y = 0, z = 0},
	[4] = {x = 128, y = -64, z = 0},
	[5] = {x = 128, y = -128, z = 0},
	[6] = {x = 64, y = -128, z = 0},
	[7] = {x = 0, y = -128, z = 0},
	[8] = {x = -64, y = -128, z = 0},
	[9] = {x = -128, y = -128, z = 0},
})

-- Small octagon
table.insert(randomringpatterns, {
	[1] = {x = 96, y = 32, z = 0},
	[2] = {x = 96, y = -32, z = 0},
	[3] = {x = -96, y = 32, z = 0},
	[4] = {x = -96, y = -32, z = 0},
	[5] = {x = 32, y = 96, z = 0},
	[6] = {x = 32, y = -96, z = 0},
	[7] = {x = -32, y = 96, z = 0},
	[8] = {x = -32, y = -96, z = 0},
})

-- Large octagon
table.insert(randomringpatterns, {
	[1] = {x = 192, y = 0, z = 0},
	[2] = {x = 192, y = 64, z = 0},
	[3] = {x = 192, y = -64, z = 0},
	[4] = {x = -192, y = 0, z = 0},
	[5] = {x = -192, y = 64, z = 0},
	[6] = {x = -192, y = -64, z = 0},
	[7] = {x = 0, y = 192, z = 0},
	[8] = {x = 0, y = -192, z = 0},
	[9] = {x = 64, y = 192, z = 0},
	[10] = {x = 64, y = -192, z = 0},
	[11] = {x = -64, y = 192, z = 0},
	[12] = {x = -64, y = -192, z = 0},
	[13] = {x = 128, y = 128, z = 0},
	[14] = {x = 128, y = -128, z = 0},
	[15] = {x = -128, y = 128, z = 0},
	[16] = {x = -128, y = -128, z = 0},
})

-- Default pattern (just a single ring)
local defaultpattern = {
	[1] = {x = 0, y = 0, z = 0},
}

local function usuableSectorSpecial(s)
	local spec = GetSecSpecial(s.special, 1)

	if (spec == 1) -- Damage
	or (spec == 2) -- Water Damage
	or (spec == 3) -- Fire Damage
	or (spec == 4) -- Electric Damage
	or (spec == 5) -- Spike Damage
	or (spec == 6) or (spec == 7) -- Death Pit
	or (spec == 8) -- Instant Kill
	--or (spec == 9) or (spec == 10) -- Ring Drain
	or (spec == 11) -- SS Damage
		return false
	end

	if (GetSecSpecial(s.special, 4) == 2) -- Exit sector
		return false
	end

	return true
end

local function newSectorInfo(sector)
	local sectorinfo = {}

	sectorinfo.sector = sector

	local i = 0
	while (sector.lines[i] and sector.lines[i].valid)
		local line = sector.lines[i]

		if (sectorinfo.xmin == nil) or (line.v1.x < sectorinfo.xmin)
			sectorinfo.xmin = line.v1.x
		end
		if (line.v2.x < sectorinfo.xmin)
			sectorinfo.xmin = line.v2.x
		end

		if (sectorinfo.xmax == nil) or (line.v1.x > sectorinfo.xmax)
			sectorinfo.xmax = line.v1.x
		end
		if (line.v2.x > sectorinfo.xmax)
			sectorinfo.xmax = line.v2.x
		end

		if (sectorinfo.ymin == nil) or (line.v1.y < sectorinfo.ymin)
			sectorinfo.ymin = line.v1.y
		end
		if (line.v2.y < sectorinfo.ymin)
			sectorinfo.ymin = line.v2.y
		end

		if (sectorinfo.ymax == nil) or (line.v1.y > sectorinfo.ymax)
			sectorinfo.ymax = line.v1.y
		end
		if (line.v2.y > sectorinfo.ymax)
			sectorinfo.ymax = line.v2.y
		end

		i = $1 + 1
	end

	return sectorinfo
end

local function enoughSpaceAboveFloor(sec, floor, ceiling, reverse)
	local space = 0

	if (reverse == true)
		space = floor - ceiling
	elseif (reverse == false)
		space = ceiling - floor
	end

	if (space < 128*FRACUNIT)
		return false
	end

	local newceiling = ceiling

	for rover in sec.ffloors()
		if not (rover.flags & FF_EXISTS)
			continue
		end

		if (rover.flags & FF_BLOCKPLAYER)
		or (rover.flags & FF_BLOCKOTHERS)
		or not (usuableSectorSpecial(rover.sector))
			if ((rover.topheight <= floor) and (reverse == false))
			or ((rover.bottomheight >= floor) and (reverse == true))
				continue
			end

			if (space <= 128*FRACUNIT)
				return false
			end

			if ((rover.bottomheight < newceiling) and (reverse == false))
				newceiling = rover.bottomheight
			elseif ((rover.topheight > newceiling) and (reverse == true))
				newceiling = rover.topheight
			end
		end
	end

	if (reverse == true)
		space = floor - newceiling
	elseif (reverse == false)
		space = newceiling - floor
	end

	return (space >= 128*FRACUNIT)
end

local function newSpot(x, y, z, flip)
	local spotinfo = {}

	spotinfo.x = x
	spotinfo.y = y
	spotinfo.z = z
	spotinfo.flip = flip

	return spotinfo
end

local PATTERNSUCCESS = 0 -- no problems!
local PATTERNFAILURE_TOOMANY = 1 -- it has too many rings
local PATTERNFAILURE_XPOS = 2 -- blocked positive X
local PATTERNFAILURE_XNEG = 4 -- blocked negative X
local PATTERNFAILURE_YPOS = 8 -- blocked positive Y
local PATTERNFAILURE_YNEG = 16 -- blocked negative Y
local PATTERNFAILURE_CENTER = 32 -- blocked center

local function validPattern(pattern, numrings, x, y, z, angle, ceiling)
	local warningflags = PATTERNSUCCESS

	if (#pattern > numrings)
		-- Not enough rings to place this pattern in the first place
		return PATTERNFAILURE_TOOMANY
	end

	-- Make a square bounding box for the pattern, make sure each corner is at a similar height
	local x1 = nil
	local x2 = nil
	local y1 = nil
	local y2 = nil

	for i = 1,#pattern
		local thisx = (pattern[i].x * cos(angle)) - (pattern[i].y * sin(angle))
		local thisy = (pattern[i].y * cos(angle)) + (pattern[i].x * sin(angle))
		
		if (x1 == nil) or (thisx < x1) then x1 = thisx end
		if (x2 == nil) or (thisx > x2) then x2 = thisx end
		if (y1 == nil) or (thisy < y1) then y1 = thisy end
		if (y2 == nil) or (thisy > y2) then y2 = thisy end
	end

	-- Buffer with a little extra space to keep it out of the walls
	x1 = $1 - (64 * FRACUNIT)
	x2 = $1 + (64 * FRACUNIT)
	y1 = $1 - (64 * FRACUNIT)
	y2 = $1 + (64 * FRACUNIT)

	for i = -1,3
		local flag = 0
		local tryx = x
		local tryy = y

		if (i == -1)
			flag = $1|PATTERNFAILURE_CENTER
		else
			if (i & 1)
				flag = $1|PATTERNFAILURE_XPOS
				tryx = $1 + x2
			else
				flag = $1|PATTERNFAILURE_XNEG
				tryx = $1 + x1
			end

			if (i > 1)
				flag = $1|PATTERNFAILURE_YPOS
				tryy = $1 + y2
			else
				flag = $1|PATTERNFAILURE_YNEG
				tryy = $1 + y1
			end
		end

		local sec = R_PointInSubsector(tryx, tryy).sector
		local validz = false

		-- Make sure to check those FOFs!!
		if (ceiling)
			if (usuableSectorSpecial(sec) and (sec.ceilingheight == z))
				validz = true
			else
				for rover in sec.ffloors()
					if (not (rover.flags & FF_EXISTS))
					or (rover.flags & FF_BUSTUP)
					or (rover.flags & FF_REVERSEPLATFORM)
						continue
					end

					if not (usuableSectorSpecial(rover.sector))
						continue
					end

					if (rover.flags & FF_BLOCKPLAYER)
					--or (rover.flags & FF_BLOCKOTHERS)
					or ((rover.flags & FF_SWIMMABLE) and (rover.flags & FF_GOOWATER))
						if (rover.bottomheight == z)
							validz = true
							break
						end
					end
				end
			end
		else
			if (usuableSectorSpecial(sec) and (sec.floorheight == z))
				validz = true
			else
				for rover in sec.ffloors()
					if (not (rover.flags & FF_EXISTS))
					or (rover.flags & FF_BUSTUP)
					or (rover.flags & FF_REVERSEPLATFORM)
						continue
					end

					if not (usuableSectorSpecial(rover.sector))
						continue
					end

					if (rover.flags & FF_BLOCKPLAYER)
					--or (rover.flags & FF_BLOCKOTHERS)
					or ((rover.flags & FF_SWIMMABLE) and (rover.flags & FF_GOOWATER))
						if (rover.topheight == z)
							validz = true
							break
						end
					end
				end
			end
		end

		if (validz == false)
			warningflags = $1|flag
		end
	end

	return warningflags
end

local function spawnRingPattern(pattern, x, y, z, ang, flip, type)
	for i = 1,#pattern
		local ringpos = pattern[i]
		local newx = x + (ringpos.x * cos(ang)) - (ringpos.y * sin(ang))
		local newy = y + (ringpos.y * cos(ang)) + (ringpos.x * sin(ang))
		local newz = z

		if (flip)
			newz = $1 - mobjinfo[type].height - 24*FRACUNIT - ringpos.z
		else
			newz = $1 + 24*FRACUNIT + ringpos.z
		end

		local newring = P_SpawnMobj(newx, newy, newz, type)

		if (flip)
			newring.eflags = $1|MFE_VERTICALFLIP
			newring.flags2 = $1|MF2_OBJECTFLIP
		end
	end
end

local spots = {}
local lastmap = nil
local lastseed = nil

addHook("MapLoad", function(map)
	if (randomizer.seed == nil or randomizer.rv == nil)
		return
	end

	if (randomizer.cvars.rings.value == 0)
		return
	end

	if (maptol & TOL_NIGHTS)
		-- NiGHTS has it's own randomization built for its maps
		return
	end

	-- Only generate a new ring spots table IF we need to.
	if (lastmap != map) or (lastseed != randomizer.seed)
		spots = {}

		-- Figure out if it's OK to put rings on the ceilings
		local allowceilings = false

		if (randomizer.cvars.rings.value == 2)
			allowceilings = false
		elseif (randomizer.cvars.rings.value == 3)
			allowceilings = true
		else
			for l in lines.iterate do
				if (l.special == 433) -- Enable/Disable Gravity Flip 
					allowceilings = true
					break
				end
			end
		end

		-- Figure out size of the map, so we can reduce how often we need to calculate things
		local mapsize = 0
		local mapminx = 0
		local mapmaxx = 0
		local mapminy = 0
		local mapmaxy = 0

		for v in vertexes.iterate do
			local x = v.x / FRACUNIT
			local y = v.y / FRACUNIT

			if (x < mapminx)
				mapminx = x
			end
			if (x > mapmaxx)
				mapmaxx = x
			end
			if (y < mapminy)
				mapminy = y
			end
			if (y > mapmaxy)
				mapmaxy = y
			end
		end

		mapsize = min(mapmaxx - mapminx, mapmaxy - mapminy)
		assert(mapsize > 0, "Map has size of 0")

		local spacing = mapsize / 20
		if (spacing < 256)
			spacing = 256
		end

		local checkx = mapminx + (spacing >> 1) - spacing
		while (checkx < mapmaxx - (spacing >> 1))
			checkx = $1 + spacing

			local checky = mapminy + (spacing >> 1) - spacing
			while (checky < mapmaxy - (spacing >> 1))
				checky = $1 + spacing

				local fixedx = checkx * FRACUNIT
				local fixedy = checky * FRACUNIT

				local sinfo = newSectorInfo(R_PointInSubsector(fixedx, fixedy).sector)

				if (fixedx < sinfo.xmin) or (fixedx > sinfo.xmax)
				or (fixedy < sinfo.ymin) or (fixedy > sinfo.ymax)
					continue
				end

				-- Calculate spots
				if (usuableSectorSpecial(sinfo.sector))
					if (sinfo.sector.floorpic != "F_SKY1")
					and (enoughSpaceAboveFloor(sinfo.sector, sinfo.sector.floorheight, sinfo.sector.ceilingheight, false))
						table.insert(spots, newSpot(fixedx, fixedy, sinfo.sector.floorheight, false))
					end

					if (allowceilings)
					and (sinfo.sector.ceilingpic != "F_SKY1")
					and (enoughSpaceAboveFloor(sinfo.sector, sinfo.sector.ceilingheight, sinfo.sector.floorheight, true))
						table.insert(spots, newSpot(fixedx, fixedy, sinfo.sector.ceilingheight, true))
					end
				end

				for rover in sinfo.sector.ffloors()
					if not (rover.flags & FF_EXISTS)
						continue
					end

					if not (usuableSectorSpecial(rover.sector))
						continue
					end

					if not ((rover.flags & FF_BLOCKPLAYER)
					--or (rover.flags & FF_BLOCKOTHERS)
					or ((rover.flags & FF_SWIMMABLE) and (rover.flags & FF_GOOWATER)))
						continue
					end
					
					if (rover.flags & FF_BUSTUP)
					or ((rover.flags & FF_CRUMBLE) and (rover.flags & FF_NORETURN))
						continue
					end

					if (rover.sector.ceilingpic != "F_SKY1")
					and (enoughSpaceAboveFloor(sinfo.sector, rover.topheight, sinfo.sector.ceilingheight, false))
						table.insert(spots, newSpot(fixedx, fixedy, rover.topheight, false))
					end

					if (allowceilings)
					and (rover.sector.floorpic != "F_SKY1")
					and (enoughSpaceAboveFloor(sinfo.sector, rover.bottomheight, sinfo.sector.floorheight, true))
						table.insert(spots, newSpot(fixedx, fixedy, rover.bottomheight, true))
					end
				end
			end
		end

		lastmap = map
		lastseed = randomizer.seed
	end

	-- Count number of rings, then remove them
	local oldrings = {}
	local ringcount = 0

	for mo in thinkers.iterate("mobj")
		if not (mo and mo.valid)
			continue
		end

		if (mo.type == MT_RING)
		or (mo.type == MT_COIN)
			table.insert(oldrings, mo)
		end
	end

	ringcount = #oldrings

	if (ringcount <= 0)
		return
	end

	for i = 1,ringcount
		if (oldrings[i] and oldrings[i].valid)
			P_RemoveMobj(oldrings[i])
		end
	end

	-- Spawn coins instead in Mario Mode
	local ringtype = MT_RING
	if (mariomode)
		ringtype = MT_COIN
	end

	local tempspots = {}
	for i = 1,#spots
		tempspots[i] = spots[i]
	end

	while (ringcount > 0)
		if (#tempspots <= 0)
			-- No more spots to use for ring placement
			break
		end

		local spotindex = (randomizer.rv % #tempspots) + 1
		local usespot = tempspots[spotindex]
		randomizer.rv = abs($1 + usespot.z)

		local originx = usespot.x
		local originy = usespot.y
		local originz = usespot.z
		local onceiling = usespot.flip

		randomizer.rv = abs($1 + (originz / FRACUNIT))
		local offsetangle = FixedAngle((randomizer.rv % 16) * ((360*FRACUNIT) / 16))
		randomizer.rv = abs($1 + (offsetangle & 2))

		local startx = originx
		local starty = originy

		local attempts = #randomringpatterns-2
		local pattern = randomringpatterns[(randomizer.rv % #randomringpatterns) + 1]

		local isPatternValid = validPattern(pattern, ringcount, originx, originy, originz, offsetangle, onceiling)
		local prevValid = isPatternValid

		while (isPatternValid != PATTERNSUCCESS)
			if (attempts <= 0)
				-- Out of attempts, just give up & place a single ring.
				pattern = defaultpattern
				break
			end

			local newpattern = false -- Decide whenever or not to try out another pattern
			
			if (isPatternValid & PATTERNFAILURE_TOOMANY) 
				-- Pattern had too many rings
				newpattern = true
			end

			if (isPatternValid & PATTERNFAILURE_CENTER)
				-- The current pattern's center was blocked entirely after nudging
				newpattern = true
			end

			if ((isPatternValid & PATTERNFAILURE_XPOS) and (isPatternValid & PATTERNFAILURE_XNEG))
			or ((isPatternValid & PATTERNFAILURE_XPOS) and (prevValid & PATTERNFAILURE_XNEG))
			or ((isPatternValid & PATTERNFAILURE_XNEG) and (prevValid & PATTERNFAILURE_XPOS))
				-- The current pattern will never have enough space on the X axis
				newpattern = true
			end

			if ((isPatternValid & PATTERNFAILURE_YPOS) and (isPatternValid & PATTERNFAILURE_YNEG))
			or ((isPatternValid & PATTERNFAILURE_YPOS) and (prevValid & PATTERNFAILURE_YNEG))
			or ((isPatternValid & PATTERNFAILURE_YNEG) and (prevValid & PATTERNFAILURE_YPOS))
				-- The current pattern will never have enough space on the Y axis
				newpattern = true
			end

			if (newpattern)
				originx = startx
				originy = starty
				randomizer.rv = $1+1
				attempts = $1-1
				pattern = randomringpatterns[(randomizer.rv % #randomringpatterns) + 1]
			else
				-- Try to nudge it away from the failing edge
				if (isPatternValid & PATTERNFAILURE_XPOS)
					originx = $1 - (64*FRACUNIT)
				elseif (isPatternValid & PATTERNFAILURE_XNEG)
					originx = $1 + (64*FRACUNIT)
				end

				if (isPatternValid & PATTERNFAILURE_YPOS)
					originy = $1 - (64*FRACUNIT)
				elseif (isPatternValid & PATTERNFAILURE_YNEG)
					originy = $1 + (64*FRACUNIT)
				end
			end

			prevValid = isPatternValid
			isPatternValid = validPattern(pattern, ringcount, originx, originy, originz, offsetangle, onceiling)
			if (newpattern)
				prevValid = isPatternValid
			end
		end

		-- EXPERIMENTAL: Try another spot if it could not find anything usable here,
		-- instead of just defaulting to a single ring.
		if (isPatternValid != PATTERNSUCCESS)
			randomizer.rv = $1+ringcount
			table.remove(tempspots, spotindex)
			continue
		end

		spawnRingPattern(pattern, originx, originy, originz, offsetangle, onceiling, ringtype)
		ringcount = $1 - #pattern

		randomizer.rv = $1+ringcount
		table.remove(tempspots, spotindex)
	end

	--[[
	if (ringcount > 0)
		print("Could not place all rings! (" + ringcount + " left)")
	end
	--]]
end)
